package l13_lambda;

import utilities.InputValue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.IntStream;

class LambdaScopeTest {

	// The body of a lambda expression has the same scope as a nested block.
	// The same rules for name conflicts and shadowing apply.
	// It is illegal to declare a parameter or a local variable in the lambda
	// that has the same name as a local variable.

	// As another consequence of the "same scope" rule, the 'this' keyword
	// in a lambda expression denotes the 'this' parameter of the method that creates the lambda.

	// A lambda expression has three ingredients:

	// 1. Parameters
	// 2. A block of code
	// 3. Values for the free variables - that is, the variables that are
	// 		not parameters and not defined inside the code

	// The lambda expression below has two free variables, 'text' and 'count'.
	// The data structure representing the lambda expression must store the
	// values for these variables.

	// The technical term for a block of code together with the values of free
	// variables is a closure. In Java, lambda expressions are closures.

	public static void main(String... args) {
		repeatMessage("Hello", 5); // Prints Hello 5 times in a separate thread
		repeatMessage("World", 3);
		printNumbers();
	}

	private static void repeatMessage(String text, int count) {
		// the lambda expression accesses the parameter variables defined
		// in the enclosing scope, not in the lambda expression itself.
		for (int i = 0; i < count; i++) {
			new Thread(() -> System.out.println(text)).start();
		}
	}

	// To ensure that the captured value is well-defined, there is an important restriction.
	// In a lambda expression, you can only reference variables whose value doesn't change.
	// This is sometimes described by saying that lambda expressions capture values, not variables.

	// lambda expression can only access local variables from an enclosing scope that are
	// effectively final. An effectively final variable is never modified — it either is
	// or could be declared as final. The same rule applies to variables captured by local
	// inner classes.

	private static void printNumbers() {
		final int n = 10;
		int loop = InputValue.getInt("loop type");

//		for (int i = 0; i < n; i++) {
			// Error—cannot capture i (changes value)
//			new Thread(() -> System.out.println("local: " + i)).start();
//		}

		switch (loop) {
			case 1 -> {
				// using an array
				for (int[] i = {0}; i[0] < n; i[0]++) {
					// Compiles, but not thread safe
					new Thread(() -> System.out.println("i[0] = " + i[0])).start();
				}
			}
			case 2 -> {
				// using reference type; cannot use Integer, since it is immutable
				// public final boolean compareAndSet(int expect, int update)
				// Atomically sets the value to the given updated value if the current value == the expected value.
				// Parameters:
				// expect - the expected value
				// update - the new value
				// Returns:
				// true if successful. False return indicates that the actual value was not equal to the expected value.
				for (AtomicInteger integer = new AtomicInteger(0);
					 !integer.compareAndSet(n, 0); integer.incrementAndGet()) {
					// Compiles, but not thread safe
					new Thread(() -> System.out.println("Atomic integer: " + integer)).start();
				}
			}
			case 3 -> {
				// using local variable defined within the loop
				for (int i = 0; i < n; i++) {
					int local = i;
					new Thread(() -> System.out.println("local: " + local)).start();
				}
			}
			case 4 -> {
				// using enhanced for loop:
				// The variable of an enhanced for loop is effectively final since its scope
				// is a single iteration. A new variable number is created in each iteration
				// and assigned the next value from the numbers array.
				int[] numbers = IntStream.rangeClosed(0, n-1).toArray();
				for (int number : numbers) {
					new Thread(() -> System.out.println("number: " + number)).start(); // OK to capture
				}
			}
		}
	}
}

